import json
import math
import random
import urllib.parse as up

from tornado import httpclient as thc, httputil as thu, gen as tg
from wcpan.logger import DEBUG, EXCEPTION, INFO, WARNING

from .util import GoogleDriveError


BACKOFF_FACTOR = 2
BACKOFF_STATUSES = ('403', '500', '502', '503', '504')


class Network(object):

    def __init__(self):
        self._access_token = None
        self._http = thc.AsyncHTTPClient()
        self._backoff_level = 0

    def set_access_token(self, token):
        self._access_token = token

    async def fetch(self, method, path, args=None, headers=None, body=None,
                    consumer=None, raise_internal_error=False):
        while True:
            await self._maybe_backoff()
            try:
                rv = await self._do_request(method, path, args, headers, body,
                                            consumer, raise_internal_error)
                return rv
            except NetworkError as e:
                if e.status != '599':
                    raise
                if e._response.raise_internal_error:
                    raise
                WARNING('wcpan.drive.google') << str(e)

    async def _do_request(self, method, path, args, headers, body, consumer,
                          raise_internal_error):
        headers = self._prepare_headers(headers)
        if args is not None:
            path = thu.url_concat(path, args)

        args = {
            'url': path,
            'method': method,
            'headers': headers,
        }
        if body is not None:
            if callable(body):
                args['body_producer'] = body
            else:
                args['body'] = body
        elif method == 'PUT':
            # for resume uploads
            args['allow_nonstandard_methods'] = True
        if consumer is not None:
            args['streaming_callback'] = consumer
        if raise_internal_error:
            # do not raise timeout from client
            args['request_timeout'] = 0.0

        request = thc.HTTPRequest(**args)
        rv = await self._http.fetch(request, raise_error=False)
        rv = Response(rv, raise_internal_error)
        rv = self._handle_status(rv)
        return rv

    def _prepare_headers(self, headers):
        h = {
            'Authorization': 'Bearer {0}'.format(self._access_token),
        }
        if headers is not None:
            h.update(headers)
        h = {k: v if isinstance(v, (bytes, str)) or v is None else str(v)
             for k, v in h.items()}
        return h

    def _handle_status(self, response):
        if backoff_needed(response):
            self._increase_backoff_level()
        else:
            self._decrease_backoff_level()

        # normal response
        if response.status[0] in ('1', '2', '3'):
            return response

        # otherwise it is an error
        raise NetworkError(response)

    def _increase_backoff_level(self):
        self._backoff_level = min(self._backoff_level + 2, 10)

    def _decrease_backoff_level(self):
        self._backoff_level = max(self._backoff_level - 1, 0)

    async def _maybe_backoff(self):
        if self._backoff_level <= 0:
            return
        seed = random.random()
        power = 2 ** self._backoff_level
        s_delay = math.floor(seed * power * BACKOFF_FACTOR)
        s_delay = min(100, s_delay)
        DEBUG('wcpan.drive.google') << 'backoff for' << s_delay
        await tg.sleep(s_delay)


class Request(object):

    def __init__(self, request):
        self._request = request

    @property
    def uri(self):
        return self._request.url

    @property
    def method(self):
        return self._request.method

    @property
    def headers(self):
        return self._request.headers


class Response(object):

    def __init__(self, response, raise_internal_error):
        self._response = response
        self._raise_internal_error = raise_internal_error
        self._request = Request(response.request)
        self._status = str(self._response.code)
        self._parsed_json = False
        self._json = None

    @property
    def status(self):
        return self._status

    @property
    def reason(self):
        return self._response.reason

    @property
    def json_(self):
        if not self._parsed_json:
            rv = self._response.body
            if rv:
                rv = rv.decode('utf-8')
                try:
                    rv = json.loads(rv)
                except ValueError as e:
                    EXCEPTION('wcpan.drive.google') << rv
                    rv = None
            self._json = rv
            self._parsed_json = True
        return self._json

    @property
    def request(self):
        return self._request

    @property
    def error(self):
        return self._response.error

    @property
    def raise_internal_error(self):
        return self._raise_internal_error

    def get_header(self, key):
        h = self._response.headers.get_list(key)
        return None if not h else h[0]


class NetworkError(GoogleDriveError):

    def __init__(self, response):
        self._response = response
        self._message = '{0} {1} - {2}'.format(self.status,
                                               self._response.reason,
                                               self.json_)

    def __str__(self):
        return self._message

    @property
    def status(self):
        return self._response.status

    @property
    def fatal(self):
        return not backoff_needed(self._response)

    @property
    def json_(self):
        return self._response.json_

    @property
    def error(self):
        return self._response.error


def initialize():
    args = {
        'max_body_size': 10 * (1024 ** 3),
    }
    thc.AsyncHTTPClient.configure(None, **args)


def backoff_needed(response):
    if response.status not in BACKOFF_STATUSES:
        return False

    # if it is not a rate limit error, it could be handled immediately
    if response.status == '403':
        msg = response.json_
        domain = msg['error']['errors'][0]['domain']
        if domain != 'usageLimits':
            return False
        INFO('wcpan.drive.google') << msg['error']['message']

    return True


initialize()



import functools as ft
import hashlib as hl
import mimetypes
import multiprocessing as mp
import os
import os.path as op
import re

from wcpan.logger import INFO, WARNING, DEBUG
import wcpan.worker as ww

from .api import Client
from .database import Database, Node
from .network import NetworkError
from .util import Settings, GoogleDriveError, stream_md5sum, FOLDER_MIME_TYPE, CHUNK_SIZE


FILE_FIELDS = 'id,name,mimeType,trashed,parents,createdTime,modifiedTime,md5Checksum,size'
CHANGE_FIELDS = 'nextPageToken,newStartPageToken,changes(fileId,removed,file({0}))'.format(FILE_FIELDS)
EMPTY_MD5SUM = 'd41d8cd98f00b204e9800998ecf8427e'


off_main_thread = ww.off_main_thread_method('_pool')


class Drive(object):

    def __init__(self, settings_path=None):
        self._settings = Settings(settings_path)
        self._client = Client(self._settings)
        self._db = Database(self._settings)
        self._pool = ww.create_thread_pool()

    def close(self):
        self._pool.shutdown()
        self._db.close()

    async def initialize(self):
        await self._client.initialize()
        self._db.initialize()

    async def sync(self):
        assert self._client.authorized

        INFO('wcpan.drive.google') << 'sync begin'

        try:
            check_point = self._db.get_metadata('check_point')
        except KeyError:
            check_point = '1'

        # first time, get root node
        if check_point == '1':
            rv = await self._client.files.get('root', fields=FILE_FIELDS)
            rv = rv.json_
            rv['name'] = None
            rv['parents'] = []
            node = Node.from_api(rv)
            self._db.insert_node(node)

        new_start_page_token = None
        changes_list_args = {
            'page_token': check_point,
            'page_size': 1000,
            'restrict_to_my_drive': True,
            'fields': CHANGE_FIELDS,
        }

        while new_start_page_token is None:
            rv = await self._client.changes.list_(**changes_list_args)
            rv = rv.json_
            next_page_token = rv.get('nextPageToken', None)
            new_start_page_token = rv.get('newStartPageToken', None)
            changes = rv['changes']

            check_point = next_page_token if next_page_token is not None else new_start_page_token

            await self._apply_changes(changes, check_point)
            changes_list_args['page_token'] = check_point

            INFO('wcpan.drive.google') << 'applied' << len(changes) << 'changes'

        INFO('wcpan.drive.google') << 'sync end'
        return True

    @property
    def root_node(self):
        return self._db.root_node

    @off_main_thread
    def get_node_by_id(self, node_id):
        return self._db.get_node_by_id(node_id)

    @off_main_thread
    def get_node_by_path(self, path):
        return self._db.get_node_by_path(path)

    @off_main_thread
    def get_path(self, node):
        return self._db.get_path_by_id(node.id_)

    @off_main_thread
    def get_path_by_id(self, node_id):
        return self._db.get_path_by_id(node_id)

    @off_main_thread
    def get_child_by_id(self, node_id, name):
        return self._db.get_child_by_id(node_id, name)

    @off_main_thread
    def get_child(self, node, name):
        return self._db.get_child_by_id(node.id_, name)

    @off_main_thread
    def get_children(self, node):
        return self._db.get_children_by_id(node.id_)

    @off_main_thread
    def get_children_by_id(self, node_id):
        return self._db.get_children_by_id(node_id)

    @off_main_thread
    def find_nodes_by_regex(self, pattern):
        return self._db.find_nodes_by_regex(pattern)

    @off_main_thread
    def find_duplicate_nodes(self):
        return self._db.find_duplicate_nodes()

    @off_main_thread
    def find_orphan_nodes(self):
        return self._db.find_orphan_nodes()

    async def download_file_by_id(self, node_id, path):
        node = await self.get_node_by_id(node_id)
        return await self.download_file(node, path)

    async def download_file(self, node, path):
        # sanity check
        if not node:
            raise ValueError('node is none')
        if node.is_folder:
            raise ValueError('node should be a file')
        if not op.isdir(path):
            raise ValueError('{0} does not exist'.format(path))

        # check if exists
        complete_path = op.join(path, node.name)
        if op.isfile(complete_path):
            return True

        # exists but not a file
        if op.exists(complete_path):
            msg = '{0} exists but is not a file'.format(complete_path)
            raise DownloadError(msg)

        # resume download
        tmp_path = complete_path + '.__tmp__'
        if op.isfile(tmp_path):
            offset = op.getsize(tmp_path)
        elif op.exists(tmp_path):
            msg = '{0} exists but is not a file'.format(complete_path)
            raise DownloadError(msg)
        else:
            offset = 0
        range_ = (offset, node.size)

        with open(tmp_path, 'ab') as fout:
            writer = ft.partial(file_consumer, fout)

            api = self._client.files
            rv = await api.download(file_id=node.id_, range_=range_,
                                    consumer=writer)

        # rename it back if completed
        os.rename(tmp_path, complete_path)

        return True

    async def create_folder(self, parent_node, folder_name):
        # sanity check
        if not parent_node:
            raise UploadError('invalid parent node')
        if not parent_node.is_folder:
            raise UploadError('invalid parent node')

        # do not create again if there is a same file
        node = await self.fetch_child_by_id(parent_node.id_, folder_name)
        if node:
            INFO('wcpan.drive.google') << 'skipped (existing)' << folder_name
            return node

        api = self._client.files
        rv = await api.create_folder(folder_name=folder_name,
                                     parent_id=parent_node.id_)
        rv = rv.json_
        node = await self.fetch_node_by_id(rv['id'])

        return node

    async def upload_file(self, file_path, parent_node, exist_ok=False):
        # sanity check
        if not parent_node:
            raise UploadError('invalid parent node')
        if not parent_node.is_folder:
            raise UploadError('invalid parent node')
        if not op.isfile(file_path):
            raise UploadError('invalid file path')

        api = self._client.files
        file_name = op.basename(file_path)

        # do not upload if remote exists a same file
        node = await self.fetch_child_by_id(parent_node.id_, file_name)
        if node:
            if exist_ok:
                INFO('wcpan.drive.google') << 'skipped (existing)' << file_path
                return node
            else:
                raise UploadConflictedError(file_path)

        total_file_size = op.getsize(file_path)
        mt, e = mimetypes.guess_type(file_path)
        if total_file_size <= 0:
            rv = await api.create_empty_file(file_name=file_name,
                                             parent_id=parent_node.id_,
                                             mime_type=mt)
            local_md5 = EMPTY_MD5SUM
        else:
            args = {
                'file_path': file_path,
                'file_name': file_name,
                'total_file_size': total_file_size,
                'parent_id': parent_node.id_,
                'mime_type': mt,
            }
            rv, local_md5 = await self._inner_upload_file(**args)

        rv = rv.json_
        node = await self.fetch_node_by_id(rv['id'])

        if node.md5 != local_md5:
            raise UploadError('md5 mismatch')

        return node

    async def fetch_child_by_id(self, node_id, name):
        safe_name = re.sub(r"[\\']", r"\\\g<0>", name)
        query = "'{0}' in parents and name = '{1}'".format(node_id,
                                                           safe_name)
        fields = 'files({0})'.format(FILE_FIELDS)
        while True:
            try:
                rv = await self._client.files.list_(q=query, fields=fields)
                break
            except NetworkError as e:
                if e.status == '400':
                    DEBUG('wcpan.drive.google') << 'failed query:' << query
                if e.fatal:
                    raise

        rv = rv.json_
        files = rv['files']
        if not files:
            return None

        node = Node.from_api(files[0])
        self._db.insert_node(node)
        return node

    async def fetch_node_by_id(self, node_id):
        rv = await self._client.files.get(node_id, fields=FILE_FIELDS)
        rv = rv.json_
        node = Node.from_api(rv)
        self._db.insert_node(node)
        return node

    async def trash_node_by_id(self, node_id):
        if node_id == self.root_node.id_:
            return
        await self._client.files.update(node_id, trashed=True)
        node = await self.get_node_by_id(node_id)
        node.is_trashed = True
        self._db.insert_node(node)
        return node

    async def trash_node(self, node):
        return await self.trash_node_by_id(node.id_)

    async def _inner_upload_file(self, file_path, file_name, total_file_size,
                                 parent_id, mime_type):
        api = self._client.files

        rv = await api.initiate_uploading(file_name=file_name,
                                          total_file_size=total_file_size,
                                          parent_id=parent_id,
                                          mime_type=mime_type)

        url = rv.get_header('Location')

        with open(file_path, 'rb') as fin:
            hasher = hl.md5()
            reader = ft.partial(file_producer, fin, hasher)
            uploader = ft.partial(self._inner_try_upload_file,
                                  url=url, producer=reader,
                                  total_file_size=total_file_size,
                                  mime_type=mime_type)

            retried = False
            offset = 0
            while True:
                ok, rv = await uploader(offset=offset)
                if ok:
                    break
                offset = rv
                fin.seek(offset, os.SEEK_SET)
                retried = True

            if retried:
                fin.seek(0, os.SEEK_SET)
                local_md5 = stream_md5sum(fin)
            else:
                local_md5 = hasher.hexdigest()

        return rv, local_md5

    async def _inner_try_upload_file(self, url, producer, offset,
                                     total_file_size, mime_type):
        api = self._client.files

        try:
            rv = await api.upload(url, producer=producer, offset=offset,
                                  total_file_size=total_file_size,
                                  mime_type=mime_type)
            return True, rv
        except NetworkError as e:
            if e.status == '404':
                raise UploadError('the upload session has been expired')
            if e.fatal:
                raise

        while True:
            try:
                rv = await api.get_upload_status(url, total_file_size)
                break
            except NetworkError as e:
                if e.fatal:
                    raise

        if rv.status != '308':
            raise UploadError('invalid upload status')
        rv = rv.get_header('Range')
        if not rv:
            raise UploadError('invalid upload status')
        rv = re.match(r'bytes=(\d+)-(\d+)', rv)
        if not rv:
            raise UploadError('invalid upload status')
        rv = int(rv.group(2))

        return False, rv

    @off_main_thread
    def _apply_changes(self, changes, check_point):
        self._db.apply_changes(changes, check_point)


class DownloadError(GoogleDriveError):

    def __init__(self, message):
        self._message = message

    def __str__(self):
        return self._message


class UploadError(GoogleDriveError):

    def __init__(self, message):
        self._message = message

    def __str__(self):
        return self._message


class UploadConflictedError(UploadError):

    def __init__(self, file_path):
        self._file_path = file_path

    def __str__(self):
        return 'remote file already exists: ' + self._file_path


async def file_producer(fin, hasher, write):
    while True:
        chunk = fin.read(CHUNK_SIZE)
        if not chunk:
            break
        hasher.update(chunk)
        await write(chunk)


def file_consumer(fout, chunk):
    fout.write(chunk)



import json
from typing import Awaitable, Callable, Dict, Tuple

from wcpan.logger import EXCEPTION, DEBUG

from .network import Network, Response, NetworkError
from .util import CommandLineOAuth2Handler, FOLDER_MIME_TYPE


API_ROOT = 'https://www.googleapis.com/drive/v3'
EMPTY_STRING = ''

Consumer = Callable[[bytes], None]
Producer = Callable[[Callable[[bytes], Awaitable[None]]], Awaitable[None]]
# producer consumer


class Client(object):

    def __init__(self, settings):
        self._settings = settings
        self._oauth = None
        self._network = None
        self._api = None

    async def initialize(self):
        if self._oauth:
            return

        s = self._settings
        args = {
            'client_config_backend': s['client_config_backend'],
            'client_config_file': s['client_config_file'],
            'get_refresh_token': s['get_refresh_token'],
            'save_credentials': s['save_credentials'],
            'save_credentials_backend': s['save_credentials_backend'],
            'save_credentials_file': s['save_credentials_file'],
        }
        self._oauth = CommandLineOAuth2Handler(args)
        await self._oauth.authorize()
        self._network = Network()
        self._network.set_access_token(self._oauth.access_token)
        self._api = {
            'changes': Changes(self),
            'files': Files(self),
        }

    @property
    def authorized(self):
        return self._oauth is not None

    @property
    def changes(self):
        return self._api['changes']

    @property
    def files(self):
        return self._api['files']

    async def _refresh_token(self):
        DEBUG('wcpan.drive.google') << 'refresh token'
        await self._oauth.refresh_access_token()
        self._network.set_access_token(self._oauth.access_token)

    async def _do_request(self, *args, **kwargs):
        while True:
            try:
                rv = await self._network.fetch(*args, **kwargs)
                break
            except NetworkError as e:
                if e.status != '401':
                    raise
            await self._refresh_token()
        return rv


class Changes(object):

    def __init__(self, client: Client) -> None:
        self._client = client
        self._root = API_ROOT + '/changes'

    async def get_start_page_token(self, supports_team_drives: bool = None,
                                   team_drive_id: str = None) -> Response:
        args = {}
        if supports_team_drives is not None:
            args['supportsTeamDrives'] = supports_team_drives
        if team_drive_id is not None:
            args['teamDriveId'] = team_drive_id

        uri = self._root + '/startPageToken'
        rv = await self._client._do_request('GET', uri, args)
        return rv

    async def list_(self, page_token: str, include_corpus_removals: bool = None,
                    include_removed: bool = None,
                    include_team_drive_items: bool = None,
                    page_size: int = None, restrict_to_my_drive: bool = None,
                    spaces: str = None, supports_team_drives: bool = None,
                    team_drive_id: str = None, fields: str = None) -> Response:
        args = {
            'pageToken': page_token,
        }
        if include_corpus_removals is not None:
            args['includeCorpusRemovals'] = include_corpus_removals
        if include_removed is not None:
            args['includeRemoved'] = include_removed
        if include_team_drive_items is not None:
            args['includeTeamDriveItems'] = include_team_drive_items
        if page_size is not None:
            args['pageSize'] = page_size
        if restrict_to_my_drive is not None:
            args['restrictToMyDrive'] = restrict_to_my_drive
        if spaces is not None:
            args['spaces'] = spaces
        if supports_team_drives is not None:
            args['supportsTeamDrives'] = supports_team_drives
        if team_drive_id is not None:
            args['teamDriveId'] = team_drive_id
        if fields is not None:
            args['fields'] = fields

        rv = await self._client._do_request('GET', self._root, args)
        return rv


class Files(object):

    def __init__(self, client: Client) -> None:
        self._client = client
        self._root = API_ROOT + '/files'
        self._upload_uri = 'https://www.googleapis.com/upload/drive/v3/files'

    # only for metadata
    async def get(self, file_id: str, supports_team_drives: bool = None,
                  fields: str = None) -> Response:
        args = {}
        if supports_team_drives is not None:
            args['supportsTeamDrives'] = supports_team_drives
        if fields is not None:
            args['fields'] = fields

        uri = self._root + '/' + file_id
        rv = await self._client._do_request('GET', uri, args)
        return rv

    async def list_(self, corpora: str = None, corpus: str = None,
                    include_team_drive_items: bool = None, order_by: str = None,
                    page_size: int = None, page_token: str = None,
                    q: str = None, spaces: str = None,
                    supports_team_drives: bool = None,
                    team_drive_id: str = None, fields: str = None) -> Response:
        args = {}
        if corpora is not None:
            args['corpora'] = corpora
        if corpus is not None:
            args['corpus'] = corpus
        if include_team_drive_items is not None:
            args['includeTeamDriveItems'] = include_team_drive_items
        if order_by is not None:
            args['orderBy'] = order_by
        if page_size is not None:
            args['pageSize'] = page_size
        if page_token is not None:
            args['pageToken'] = page_token
        if q is not None:
            args['q'] = q
        if spaces is not None:
            args['spaces'] = spaces
        if supports_team_drives is not None:
            args['supportsTeamDrives'] = supports_team_drives
        if team_drive_id is not None:
            args['teamDriveId'] = team_drive_id
        if fields is not None:
            args['fields'] = fields

        rv = await self._client._do_request('GET', self._root, args)
        return rv

    # download and send to streaming_callback
    async def download(self, file_id: str, range_: Tuple[int, int],
                       consumer: Consumer, acknowledge_abuse: bool = None,
                       supports_team_drives: bool = None) -> Response:
        args = {
            'alt': 'media',
        }
        if acknowledge_abuse is not None:
            args['acknowledgeAbuse'] = acknowledge_abuse
        if supports_team_drives is not None:
            args['supportsTeamDrives'] = supports_team_drives

        headers = {
            'Range': 'bytes={0}-{1}'.format(*range_),
        }

        uri = self._root + '/' + file_id
        rv = await self._client._do_request('GET', uri, args, headers=headers,
                                            consumer=consumer,
                                            raise_internal_error=True)
        return rv

    async def initiate_uploading(self, file_name: str, total_file_size: int,
                                 parent_id: str = None,
                                 mime_type: str = None) -> Response:
        if not file_name:
            raise ValueError('file name is empty')
        if total_file_size <= 0:
            raise ValueError('please use create_empty_file() to create an empty file')

        metadata = {
            'name': file_name,
        }
        if parent_id is not None:
            metadata['parents'] = [parent_id]
        metadata = json.dumps(metadata)
        metadata = metadata.encode('utf-8')
        headers = {
            'X-Upload-Content-Length': total_file_size,
            'Content-Type': 'application/json; charset=UTF-8',
            'Content-Length': len(metadata),
        }
        if mime_type is not None:
            headers['X-Upload-Content-Type'] = mime_type

        args = {
            'uploadType': 'resumable',
        }

        rv = await self._client._do_request('POST', self._upload_uri, args,
                                            headers=headers, body=metadata)
        return rv

    async def upload(self, uri: str, producer: Producer, offset: int,
                     total_file_size: int, mime_type: str = None) -> Response:
        if not uri:
            raise ValueError('invalid session URI')
        if not producer:
            raise ValueError('invalid body producer')
        if total_file_size <= 0:
            raise ValueError('please use create_empty_file() to create an empty file')
        if offset < 0 or offset >= total_file_size:
            raise ValueError('offset is out of range')

        last_position = total_file_size - 1
        headers = {
            'Content-Length': total_file_size - offset,
            'Content-Range': 'bytes {0}-{1}/{2}'.format(offset, last_position,
                                                        total_file_size),
        }
        if mime_type is not None:
            headers['Content-Type'] = mime_type

        # upload usually need more time to complete
        rv = await self._client._do_request('PUT', uri, headers=headers,
                                            body=producer,
                                            raise_internal_error=True)
        return rv

    async def get_upload_status(self, uri: str,
                                total_file_size: int) -> Response:
        if not uri:
            raise ValueError('invalid session URI')
        if total_file_size <= 0:
            raise ValueError('please use create_empty_file() to create an empty file')

        headers = {
            'Content-Length': 0,
            'Content-Range': 'bytes */{0}'.format(total_file_size),
        }
        rv = await self._client._do_request('PUT', uri, headers=headers)
        return rv

    async def create_folder(self, folder_name: str,
                            parent_id: str = None) -> Response:
        metadata = {
            'name': folder_name,
            'mimeType': FOLDER_MIME_TYPE,
        }
        if parent_id is not None:
            metadata['parents'] = [parent_id]
        metadata = json.dumps(metadata)
        metadata = metadata.encode('utf-8')
        headers = {
            'Content-Type': 'application/json; charset=UTF-8',
            'Content-Length': len(metadata),
        }

        rv = await self._client._do_request('POST', self._root, headers=headers,
                                            body=metadata)
        return rv

    async def create_empty_file(self, file_name: str,
                                parent_id: str = None,
                                mime_type: str = None) -> Response:
        if not file_name:
            raise ValueError('file name is empty')

        metadata = {
            'name': file_name,
        }
        if parent_id is not None:
            metadata['parents'] = [parent_id]
        if mime_type is not None:
            metadata['mimeType'] = mime_type
        else:
            metadata['mimeType'] = 'application/octet-stream'
        metadata = json.dumps(metadata)
        metadata = metadata.encode('utf-8')
        headers = {
            'Content-Type': 'application/json; charset=UTF-8',
            'Content-Length': len(metadata),
        }

        rv = await self._client._do_request('POST', self._root, headers=headers,
                                            body=metadata)
        return rv

    async def update(self, file_id: str, trashed: bool = None) -> Response:
        metadata = {}
        if trashed is not None:
            metadata['trashed'] = trashed

        if not metadata:
            raise ValueError('not enough parameter')

        metadata = json.dumps(metadata)
        metadata = metadata.encode('utf-8')
        headers = {
            'Content-Type': 'application/json; charset=UTF-8',
            'Content-Length': len(metadata),
        }

        uri = self._root + '/' + file_id
        rv = await self._client._do_request('PATCH', uri, headers=headers,
                                            body=metadata)
        return rv



import datetime as dt
import functools as ft
import hashlib
import json
import os
import os.path as op
import re
import urllib.parse as up

from tornado import auth as ta
import yaml


ISO_PATTERN = r'^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(.(\d{3,6}))?(Z|(([+\-])(\d{2}):(\d{2})))$'
CHUNK_SIZE = 64 * 1024
FOLDER_MIME_TYPE = 'application/vnd.google-apps.folder'


class GoogleDriveError(Exception):
    pass


class InvalidDateTimeError(GoogleDriveError):

    def __init__(self, iso):
        self._iso = iso

    def __str__(self):
        return 'invalid ISO date: ' + self._iso


class Settings(object):

    def __init__(self, path):
        self._path = path
        self._data = {
            'version': 1,
            'save_credentials': False,
            'client_config_backend': 'settings',
            'nodes_database_file': ':memory:',
        }
        self._initialize()

    def _initialize(self):
        if not self._path:
            return

        # default values for file
        self._data['version'] = 1
        self._data['client_config_backend'] = 'file'
        self._data['client_config_file'] = op.join(self._path, 'client_secret.json')
        self._data['get_refresh_token'] = True
        self._data['save_credentials'] = True
        self._data['save_credentials_backend'] = 'file'
        self._data['save_credentials_file'] = op.join(self._path, 'oauth_token.yaml')
        self._data['nodes_database_file'] = op.join(self._path, 'nodes.db')

        os.makedirs(self._path, exist_ok=True)
        path = op.join(self._path, 'settings.yaml')
        if not op.exists(path):
            rv = yaml.dump(self._data, default_flow_style=False)
            with open(path, 'w') as fout:
                fout.write(rv)
        else:
            with open(path, 'r') as fin:
                rv = yaml.load(fin)
            self._data.update(rv)

    def __getitem__(self, key):
        return self._data[key]


class CommandLineOAuth2Handler(ta.GoogleOAuth2Mixin):

    def __init__(self, cfg):
        super(CommandLineOAuth2Handler, self).__init__()

        self._load(cfg)
        self._code = ''
        self._settings = {
            'google_oauth': {
                'key': self._client_id,
                'secret': self._client_secret,
            },
        }

    async def authorize(self):
        if self._access_token is None:
            await self._get_access_token()

    async def refresh_access_token(self):
        curl = self.get_auth_http_client()
        body = up.urlencode({
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'refresh_token': self._refresh_token,
            'grant_type': 'refresh_token',
        })
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        response = await curl.fetch(self._OAUTH_ACCESS_TOKEN_URL,
                                    method='POST', headers=headers, body=body)
        if response.error:
            raise ValueError('Google OAuth 2.0 error: {0}'.format(response))
        token = json.loads(response.body)
        self._save_token(token)

    async def _get_access_token(self):
        await self.authorize_redirect(redirect_uri=self._redirect_uri,
                                      scope=['https://www.googleapis.com/auth/drive'],
                                      client_id=self._client_id)
        token = await self.get_authenticated_user(self._redirect_uri, code=self._code)
        self._code = ''
        self._save_token(token)

    @property
    def access_token(self):
        assert self._access_token is not None
        return self._access_token

    def redirect(self, url):
        print(url)
        self._code = input()

    @property
    def settings(self):
        return self._settings

    def _load(self, cfg):
        if 'client_config_file' not in cfg:
            raise ValueError('`client_config_file` not found')

        # load API key
        with open(cfg['client_config_file'], 'r') as fin:
            client = json.load(fin)
        if 'installed' not in client:
            raise ValueError('credential should be an installed application')
        client = client['installed']
        self._redirect_uri = client['redirect_uris'][0]
        self._client_id = client['client_id']
        self._client_secret = client['client_secret']

        # load refresh token
        self._token_path = cfg['save_credentials_file']
        if not op.isfile(self._token_path):
            self._access_token = None
            self._refresh_token = None
            return
        with open(self._token_path, 'r') as fin:
            token = yaml.safe_load(fin)
        if token.get('version', 0) != 1:
            raise ValueError('wrong token file')
        self._access_token = token['access_token']
        self._refresh_token = token['refresh_token']

    def _save_token(self, token):
        self._access_token = token['access_token']
        if 'refresh_token' in token:
            self._refresh_token = token['refresh_token']
        token = {
            'version': 1,
            'access_token': self._access_token,
            'refresh_token': self._refresh_token,
        }

        # save refresh token
        rv = yaml.dump(token, default_flow_style=False)
        with open(self._token_path, 'w') as fout:
            fout.write(rv)


def from_isoformat(iso_datetime):
    rv = re.match(ISO_PATTERN, iso_datetime)
    if not rv:
        raise InvalidDateTimeError(iso_datetime)

    year = int(rv.group(1), 10)
    month = int(rv.group(2), 10)
    day = int(rv.group(3), 10)
    hour = int(rv.group(4), 10)
    minute = int(rv.group(5), 10)
    second = int(rv.group(6), 10)
    if rv.group(8):
        microsecond = rv.group(8).ljust(6, '0')
        microsecond = int(microsecond, 10)
    else:
        microsecond = 0
    tz = rv.group(9)
    if tz == 'Z':
        tz = dt.timezone.utc
    else:
        f = rv.group(11)
        h = int(rv.group(12), 10)
        m = int(rv.group(13), 10)
        tz = dt.timedelta(hours=h, minutes=m)
        if f == '-':
            tz = -tz
        tz = dt.timezone(tz)

    rv = dt.datetime(year, month, day, hour, minute, second, microsecond, tz)
    return rv


def stream_md5sum(input_stream):
    hasher = hashlib.md5()
    while True:
        chunk = input_stream.read(CHUNK_SIZE)
        if not chunk:
            break
        hasher.update(chunk)
    return hasher.hexdigest()
